// Author: Jean Honorio
// E-mail: jhonorio@csail.mit.edu

#include "mex.h"
#include <math.h>
#include <stdlib.h>

void mexFunction(int nargout, mxArray *argout[], int nargin, const mxArray *argin[])
{
    int frequency, derivative, N, *x, r, i, j, k, l;
    long power_2_N;
    double *W, *b, *W_ptr, A, exp_fx, Z, *DW, *Db, *table, *table_ptr;

    if (nargin != 2)
        mexErrMsgTxt("Usage: log_partition_ising(W,b);");
	frequency = (nargout == 2);
    derivative = (nargout == 3);
    if (!frequency && !derivative && (nargout > 1))
        mexErrMsgTxt("Usage: logZ = log_partition_ising(W,b);\n   [logZ table] = log_partition_ising(W,b);\n   [logZ DW Db] = log_partition_ising(W,b);");
    N = mxGetM(argin [0]);
    if (mxGetN(argin [0]) != N)
        mexErrMsgTxt("Matrix W is not square");
    W = mxGetPr(argin [0]);
    if (mxGetM(argin [1]) * mxGetN(argin [1]) != N)
        mexErrMsgTxt("Vector b is not the same dimension of W");
    b = mxGetPr(argin [1]);
    if ((x = (int *) calloc(N, sizeof(int))) == NULL)
        mexErrMsgTxt("Insufficient memory");
	if (frequency) {
		power_2_N = 1<<N;
		argout [1] = mxCreateDoubleMatrix(1, power_2_N, mxREAL);
		table = mxGetPr(argout[1]);
	} else
		table = NULL;
	if (derivative) {
		argout [1] = mxCreateDoubleMatrix(N, N, mxREAL);
		DW = mxGetPr(argout[1]);
		argout [2] = mxCreateDoubleMatrix(N, 1, mxREAL);
		Db = mxGetPr(argout[2]);
	} else {
		DW = NULL;
		Db = NULL;
	}
    W_ptr = W;
    A = 0;
    for (i = 0; i < N; i++) {
        double s = 0;
        for (j = 0; j < N; j++)
            s += fabs(*(W_ptr++));
        A += s / 2 + fabs(b [i]);
    }
    for (r = 0; r < N; r++)
        x [r] = -3;
    r = 0;
    Z = 0;
	if (frequency)
		table_ptr = table;
    while (r >= 0) {
        int *x_r = x + r;
        *x_r += 2;
        if (*x_r > 1)
            *x_r = -3;
        if (*x_r == -3)
            r--;
        else if (r == N - 1) {
            double a = 0;
            double *W_ptr = W;
            for (i = 0; i < N; i++) {
                double s = 0;
                int *x_ptr = x;
                for (j = 0; j < N; j++)
                    s += *(W_ptr++) * *(x_ptr++);
                a += x [i] * (s / 2 + b [i]);
            }
			exp_fx = exp(a - A);
            Z += exp_fx;
			if (frequency)
				*(table_ptr++) = exp_fx;
			else if (derivative) {
				double *DW_ptr = DW;
				for (k = 0; k < N; k++) {
					int *x_ptr = x;
					for (l = 0; l < N; l++)
						*(DW_ptr++) += exp_fx * (x [k] * *(x_ptr++) - (k == l)) / 2;
					Db [k] += exp_fx * x [k];
				}
			}
        } else
            r++;
    }
    free(x);
	argout [0] = mxCreateDoubleScalar(A + log(Z));
	if (frequency) {
	    table_ptr = table;
	    for (k = 0; k < power_2_N; k++)
	        *(table_ptr++) /= Z;
	} else if (derivative) {
		double *DW_ptr = DW;
		for (k = 0; k < N; k++) {
			for (l = 0; l < N; l++)
				*(DW_ptr++) /= Z;
			Db [k] /= Z;
		}
	}
}
